////////////////////////////////////////////////////////////////////////////////////////////////////
// test_fastest_sort_w4.cpp
// Copyright(c) 2018 ozk
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "fastest_sort_w4.h"

#include <algorithm>
#include <array>
#include <chrono>
#include <iostream>
#include <list>
#include <string>
#include <vector>


namespace fswn = fsw4;
using fastest_sort_wn = fsw4::fastest_sort_w4<int>;


namespace
{

	const size_t FSW_N = 4;

	template<typename Iterator>
	void generate_data(const Iterator &first, const Iterator &last)
	{
		using Value = typename std::iterator_traits<Iterator>::value_type;
		static const Value data[FSW_N] = { 3, 4, 1, 2 };
		int k = 0;
		for (Iterator i = first; i != last && k < FSW_N; ++i, ++k) {
			*i = data[k];
		}
	}

	template<size_t N>
	void generate_permutation(std::vector<std::array<int, N>> &permutation)
	{
		std::array<int, N> a;
		for (size_t i = 0; i < N; ++i) {
			a[i] = int(i);
		}
		do {
			permutation.push_back(a);
		} while (std::next_permutation(a.begin(), a.end()));
	}

	struct MyData1
	{
		std::string key;
	};

	struct MyCompare1
	{
		bool operator()(const MyData1 &lhs, const MyData1 &rhs) const
		{
			return lhs.key > rhs.key;
		}
	};

	bool operator==(const MyData1 &lhs, const MyData1 &rhs)
	{
		return lhs.key == rhs.key;
	}

	bool operator<(const MyData1 &lhs, const MyData1 &rhs)
	{
		return lhs.key < rhs.key;
	}

	std::ostream &operator<<(std::ostream &os, const MyData1 &x)
	{
		return os << x.key;
	}

	struct MyData2
	{
		int key;
		int order;
	};

	bool operator==(const MyData2 &lhs, const MyData2 &rhs)
	{
		return lhs.key == rhs.key && lhs.order == rhs.order;
	}

	bool operator<(const MyData2 &lhs, const MyData2 &rhs)
	{
		return lhs.key < rhs.key;
	}

	std::ostream &operator<<(std::ostream &os, const MyData2 &x)
	{
		return os << "(" << x.key << ", " << x.order << ")";
	}

	template<typename T, size_t N>
	std::ostream &operator<<(std::ostream &os, const std::array<T, N> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	template<typename T>
	std::ostream &operator<<(std::ostream &os, const std::vector<T> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	template<typename T>
	std::ostream &operator<<(std::ostream &os, const std::list<T> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	std::ostream &operator<<(std::ostream &os, const int(&x)[FSW_N])
	{
		os << "{ ";
		const char *sep = "";
		for (auto i = 0; i < FSW_N; ++i) {
			os << sep << x[i];
			sep = ", ";
		}
		os << " }";
		return os;
	}

	template<typename Container>
	bool AreContainerEqual(const Container &expected, const Container &actual)
	{
		if (expected.size() != actual.size()) {
			return false;
		}
		auto i_expected = expected.begin();
		auto i_actual = actual.begin();
		for (; i_expected != expected.end() && i_actual != actual.end(); ++i_expected, ++i_actual) {
			if (!(*i_expected == *i_actual)) {
				return false;
			}
		}
		return true;
	}

	bool AreArrayEqual(const int(&expected)[FSW_N], const int(&actual)[FSW_N])
	{
		for (size_t i = 0; i < FSW_N; ++i) {
			if (expected[i] != actual[i]) {
				return false;
			}
		}
		return true;
	}

} // unnamed


int main()
{
	{
		// max_size()
		static_assert(fastest_sort_wn::max_size() == FSW_N, "max_size() must be equal to N.");
		std::cout << "max_size() OK." << std::endl;
	} {
		// sort2(a, b)
		using array_int2_type = std::array<int, 2>;
		std::vector<array_int2_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			fswn::sort2(a, b);
			array_int2_type actual{ a, b };

			array_int2_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort2(a, b) OK." << std::endl;
	} {
		// sort3(a, b, c)
		using array_int3_type = std::array<int, 3>;
		std::vector<array_int3_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			fswn::sort3(a, b, c);
			array_int3_type actual{ a, b, c };

			array_int3_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort3(a, b, c) OK." << std::endl;
	} {
		// sort4(a, b, c, d)
		using array_int4_type = std::array<int, 4>;
		std::vector<array_int4_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			fswn::sort4(a, b, c, d);
			array_int4_type actual{ a, b, c, d };

			array_int4_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort4(a, b, c, d) OK." << std::endl;
	} {
		// sort(first, last) int[N]
		int data[FSW_N];
		generate_data(&data[0], &data[0] + FSW_N);

		int actual[FSW_N];
		for (int i = 0; i < FSW_N; ++i) {
			actual[i] = data[i];
		}
		fswn::sort(&actual[0], &actual[0] + FSW_N);

		int expected[FSW_N];
		for (int i = 0; i < FSW_N; ++i) {
			expected[i] = data[i];
		}
		std::sort(&expected[0], &expected[0] + FSW_N);

		//std::cout << data << std::endl;
		for (size_t i = 0; i < 4; ++i) {
			if (!AreArrayEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort(first, last) int[N] OK." << std::endl;
	} {
		// sort(first, last) std::vector<int>
		std::vector<int> data(FSW_N);
		generate_data(data.begin(), data.end());

		std::vector<int> actual{ data };
		fswn::sort(actual.begin(), actual.end());

		std::vector<int> expected{ data };
		std::sort(expected.begin(), expected.end());

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::vector<int> OK." << std::endl;
	} {
		// sort(first, last) std::list<int>
		std::list<int> data(FSW_N);
		generate_data(data.begin(), data.end());

		std::list<int> actual(data);
		fswn::sort(actual.begin(), actual.end());

		std::list<int> expected{ data };
		expected.sort();

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::list<int> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 0> MyCompare1
		const std::array<MyData1, 0> data{};
		const std::array<MyData1, 0> expected{};

		std::array<MyData1, 0> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 0> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 1> MyCompare1
		const std::array<MyData1, 1> data{ "a" };
		const std::array<MyData1, 1> expected{ "a" };

		std::array<MyData1, 1> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 1> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 2> MyCompare1
		const std::array<MyData1, 2> data{ "a", "b" };
		const std::array<MyData1, 2> expected{ "b", "a" };

		std::array<MyData1, 2> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 2> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 3> MyCompare1
		const std::array<MyData1, 3> data{ "c", "a", "b" };
		const std::array<MyData1, 3> expected{ "c", "b", "a" };

		std::array<MyData1, 3> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 3> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 4> MyCompare1
		const std::array<MyData1, 4> data{ "c", "a", "d", "b" };
		const std::array<MyData1, 4> expected{ "d", "c", "b", "a" };

		std::array<MyData1, 4> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 4> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 0>
		const size_t N = 0;
		std::array<MyData2, N> data{};
		std::array<MyData2, N> actual{};
		fswn::sort(actual.begin(), actual.end());

		std::array<MyData2, N> expected{};
		std::stable_sort(expected.begin(), expected.end());

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData2, 0> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 1>
		const size_t N = 1;
		for (int v1 = 1; v1 <= N; ++v1) {
			const std::array<MyData2, N> data{
				MyData2{ v1, 1 }
			};

			std::array<MyData2, N> actual(data);
			fswn::sort(actual.begin(), actual.end());

			std::array<MyData2, N> expected(data);
			std::stable_sort(expected.begin(), expected.end());

			//std::cout << data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 1> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 2>
		const size_t N = 2;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				const std::array<MyData2, N> data{
					MyData2{ v1, 1 },
					MyData2{ v2, 2 }
				};

				std::array<MyData2, N> actual(data);
				fswn::sort(actual.begin(), actual.end());

				std::array<MyData2, N> expected(data);
				std::stable_sort(expected.begin(), expected.end());

				//std::cout << data << std::endl;
				if (!AreContainerEqual(expected, actual)) {
					std::cout << "ERROR expected != actual" << std::endl;
					std::cout << "input = " << data << std::endl;
					std::cout << "expected = " << expected << std::endl;
					std::cout << "actual = " << actual << std::endl;
					return 1;
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 2> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 3>
		const size_t N = 3;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					const std::array<MyData2, N> data{
						MyData2{ v1, 1 },
						MyData2{ v2, 2 },
						MyData2{ v3, 3 }
					};

					std::array<MyData2, N> actual(data);
					fswn::sort(actual.begin(), actual.end());

					std::array<MyData2, N> expected(data);
					std::stable_sort(expected.begin(), expected.end());

					//std::cout << data << std::endl;
					if (!AreContainerEqual(expected, actual)) {
						std::cout << "ERROR expected != actual" << std::endl;
						std::cout << "input = " << data << std::endl;
						std::cout << "expected = " << expected << std::endl;
						std::cout << "actual = " << actual << std::endl;
						return 1;
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 3> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 4>
		const size_t N = 4;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						const std::array<MyData2, N> data{
							MyData2{ v1, 1 },
							MyData2{ v2, 2 },
							MyData2{ v3, 3 },
							MyData2{ v4, 4 }
						};

						std::array<MyData2, N> actual(data);
						fswn::sort(actual.begin(), actual.end());

						std::array<MyData2, N> expected(data);
						std::stable_sort(expected.begin(), expected.end());

						//std::cout << data << std::endl;
						if (!AreContainerEqual(expected, actual)) {
							std::cout << "ERROR expected != actual" << std::endl;
							std::cout << "input = " << data << std::endl;
							std::cout << "expected = " << expected << std::endl;
							std::cout << "actual = " << actual << std::endl;
							return 1;
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 4> OK." << std::endl;
	} {
		using std::chrono::steady_clock;
		using std::chrono::microseconds;
		using std::chrono::duration_cast;

		// performance
		const int LOOP = 100000;
		{
			// std::sort() N = 2;
			const size_t N = 2;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						std::array<int, N> data{ v1, v2 };
						std::sort(data.begin(), data.end());
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 2;
			const size_t N = 2;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						std::array<int, N> data{ v1, v2 };
						fswn::sort(data.begin(), data.end());
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 3;
			const size_t N = 3;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							std::array<int, N> data{ v1, v2, v3 };
							std::sort(data.begin(), data.end());
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 3;
			const size_t N = 3;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							std::array<int, N> data{ v1, v2, v3 };
							fswn::sort(data.begin(), data.end());
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 4;
			const size_t N = 4;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								std::array<int, N> data{ v1, v2, v3, v4 };
								std::sort(data.begin(), data.end());
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 4;
			const size_t N = 4;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								std::array<int, N> data{ v1, v2, v3, v4 };
								fswn::sort(data.begin(), data.end());
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		}
	}

	return 0;
}
