////////////////////////////////////////////////////////////////////////////////////////////////////
// test_fastest_sort_w8.cpp
// Copyright(c) 2018 ozk
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "fastest_sort_w8.h"

#include <algorithm>
#include <array>
#include <chrono>
#include <iostream>
#include <list>
#include <string>
#include <vector>


namespace fswn = fsw8;
using fastest_sort_wn = fsw8::fastest_sort_w8<int>;


namespace
{

	const size_t FSW_N = 8;

	template<typename Iterator>
	void generate_data(const Iterator &first, const Iterator &last)
	{
		using Value = typename std::iterator_traits<Iterator>::value_type;
		static const Value data[FSW_N] = { 4, 8, 3, 7, 2, 6, 1, 5 };
		int k = 0;
		for (Iterator i = first; i != last && k < FSW_N; ++i, ++k) {
			*i = data[k];
		}
	}

	template<size_t N>
	void generate_permutation(std::vector<std::array<int, N>> &permutation)
	{
		std::array<int, N> a;
		for (size_t i = 0; i < N; ++i) {
			a[i] = int(i);
		}
		do {
			permutation.push_back(a);
		} while (std::next_permutation(a.begin(), a.end()));
	}

	struct MyData1
	{
		std::string key;
	};

	struct MyCompare1
	{
		bool operator()(const MyData1 &lhs, const MyData1 &rhs) const
		{
			return lhs.key > rhs.key;
		}
	};

	bool operator==(const MyData1 &lhs, const MyData1 &rhs)
	{
		return lhs.key == rhs.key;
	}

	bool operator<(const MyData1 &lhs, const MyData1 &rhs)
	{
		return lhs.key < rhs.key;
	}

	std::ostream &operator<<(std::ostream &os, const MyData1 &x)
	{
		return os << x.key;
	}

	struct MyData2
	{
		int key;
		int order;
	};

	bool operator==(const MyData2 &lhs, const MyData2 &rhs)
	{
		return lhs.key == rhs.key && lhs.order == rhs.order;
	}

	bool operator<(const MyData2 &lhs, const MyData2 &rhs)
	{
		return lhs.key < rhs.key;
	}

	std::ostream &operator<<(std::ostream &os, const MyData2 &x)
	{
		return os << "(" << x.key << ", " << x.order << ")";
	}

	template<typename T, size_t N>
	std::ostream &operator<<(std::ostream &os, const std::array<T, N> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	template<typename T>
	std::ostream &operator<<(std::ostream &os, const std::vector<T> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	template<typename T>
	std::ostream &operator<<(std::ostream &os, const std::list<T> &x)
	{
		if (x.empty()) {
			os << "{}";
		}
		else {
			os << "{ ";
			const char *sep = "";
			for (auto i = x.begin(); i != x.end(); ++i) {
				os << sep << *i;
				sep = ", ";
			}
			os << " }";
		}
		return os;
	}

	std::ostream &operator<<(std::ostream &os, const int(&x)[FSW_N])
	{
		os << "{ ";
		const char *sep = "";
		for (auto i = 0; i < FSW_N; ++i) {
			os << sep << x[i];
			sep = ", ";
		}
		os << " }";
		return os;
	}

	template<typename Container>
	bool AreContainerEqual(const Container &expected, const Container &actual)
	{
		if (expected.size() != actual.size()) {
			return false;
		}
		auto i_expected = expected.begin();
		auto i_actual = actual.begin();
		for (; i_expected != expected.end() && i_actual != actual.end(); ++i_expected, ++i_actual) {
			if (!(*i_expected == *i_actual)) {
				return false;
			}
		}
		return true;
	}

	bool AreArrayEqual(const int(&expected)[FSW_N], const int(&actual)[FSW_N])
	{
		for (size_t i = 0; i < FSW_N; ++i) {
			if (expected[i] != actual[i]) {
				return false;
			}
		}
		return true;
	}

} // unnamed


int main()
{
	{
		// max_size()
		static_assert(fastest_sort_wn::max_size() == FSW_N, "max_size() must be equal to N.");
		std::cout << "max_size() OK." << std::endl;
	} {
		// sort2(a, b)
		using array_int2_type = std::array<int, 2>;
		std::vector<array_int2_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			fswn::sort2(a, b);
			array_int2_type actual{ a, b };

			array_int2_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort2(a, b) OK." << std::endl;
	} {
		// sort3(a, b, c)
		using array_int3_type = std::array<int, 3>;
		std::vector<array_int3_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			fswn::sort3(a, b, c);
			array_int3_type actual{ a, b, c };

			array_int3_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort3(a, b, c) OK." << std::endl;
	} {
		// sort4(a, b, c, d)
		using array_int4_type = std::array<int, 4>;
		std::vector<array_int4_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			fswn::sort4(a, b, c, d);
			array_int4_type actual{ a, b, c, d };

			array_int4_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort4(a, b, c, d) OK." << std::endl;
	} {
		// sort5(a, b, c, d, e)
		using array_int5_type = std::array<int, 5>;
		std::vector<array_int5_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			int e = i_data->at(4);
			fswn::sort5(a, b, c, d, e);
			array_int5_type actual{ a, b, c, d, e };

			array_int5_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort5(a, b, c, d, e) OK." << std::endl;
	} {
		// sort6(a, b, c, d, e, f)
		using array_int6_type = std::array<int, 6>;
		std::vector<array_int6_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			int e = i_data->at(4);
			int f = i_data->at(5);
			fswn::sort6(a, b, c, d, e, f);
			array_int6_type actual{ a, b, c, d, e, f };

			array_int6_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort6(a, b, c, d, e, f) OK." << std::endl;
	} {
		// sort7(a, b, c, d, e, f, g)
		using array_int7_type = std::array<int, 7>;
		std::vector<array_int7_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			int e = i_data->at(4);
			int f = i_data->at(5);
			int g = i_data->at(6);
			fswn::sort7(a, b, c, d, e, f, g);
			array_int7_type actual{ a, b, c, d, e, f, g };

			array_int7_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort7(a, b, c, d, e, f, g) OK." << std::endl;
	} {
		// sort8(a, b, c, d, e, f, g, h)
		using array_int8_type = std::array<int, 8>;
		std::vector<array_int8_type> data;
		generate_permutation(data);

		for (auto i_data = data.begin(); i_data != data.end(); ++i_data) {
			int a = i_data->at(0);
			int b = i_data->at(1);
			int c = i_data->at(2);
			int d = i_data->at(3);
			int e = i_data->at(4);
			int f = i_data->at(5);
			int g = i_data->at(6);
			int h = i_data->at(7);
			fswn::sort8(a, b, c, d, e, f, g, h);
			array_int8_type actual{ a, b, c, d, e, f, g, h };

			array_int8_type expected(*i_data);
			std::sort(expected.begin(), expected.end());

			//std::cout << *i_data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << *i_data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort8(a, b, c, d, e, f, g, h) OK." << std::endl;
	} {
		// sort(first, last) int[N]
		int data[FSW_N];
		generate_data(&data[0], &data[0] + FSW_N);

		int actual[FSW_N];
		for (int i = 0; i < FSW_N; ++i) {
			actual[i] = data[i];
		}
		fswn::sort(&actual[0], &actual[0] + FSW_N);

		int expected[FSW_N];
		for (int i = 0; i < FSW_N; ++i) {
			expected[i] = data[i];
		}
		std::sort(&expected[0], &expected[0] + FSW_N);

		//std::cout << data << std::endl;
		for (size_t i = 0; i < 4; ++i) {
			if (!AreArrayEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort(first, last) int[N] OK." << std::endl;
	} {
		// sort(first, last) std::vector<int>
		std::vector<int> data(FSW_N);
		generate_data(data.begin(), data.end());

		std::vector<int> actual{ data };
		fswn::sort(actual.begin(), actual.end());

		std::vector<int> expected{ data };
		std::sort(expected.begin(), expected.end());

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::vector<int> OK." << std::endl;
	} {
		// sort(first, last) std::list<int>
		std::list<int> data(FSW_N);
		generate_data(data.begin(), data.end());

		std::list<int> actual(data);
		fswn::sort(actual.begin(), actual.end());

		std::list<int> expected{ data };
		expected.sort();

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::list<int> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 0> MyCompare1
		const std::array<MyData1, 0> data{};
		const std::array<MyData1, 0> expected{};

		std::array<MyData1, 0> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 0> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 1> MyCompare1
		const std::array<MyData1, 1> data{ "a" };
		const std::array<MyData1, 1> expected{ "a" };

		std::array<MyData1, 1> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 1> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 2> MyCompare1
		const std::array<MyData1, 2> data{ "a", "b" };
		const std::array<MyData1, 2> expected{ "b", "a" };

		std::array<MyData1, 2> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 2> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 3> MyCompare1
		const std::array<MyData1, 3> data{ "c", "a", "b" };
		const std::array<MyData1, 3> expected{ "c", "b", "a" };

		std::array<MyData1, 3> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 3> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 4> MyCompare1
		const std::array<MyData1, 4> data{ "c", "a", "d", "b" };
		const std::array<MyData1, 4> expected{ "d", "c", "b", "a" };

		std::array<MyData1, 4> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 4> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 5> MyCompare1
		const std::array<MyData1, 5> data{ "c", "a", "e", "d", "b" };
		const std::array<MyData1, 5> expected{ "e", "d", "c", "b", "a" };

		std::array<MyData1, 5> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 5> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 6> MyCompare1
		const std::array<MyData1, 6> data{ "c", "a", "f", "e", "d", "b" };
		const std::array<MyData1, 6> expected{ "f", "e", "d", "c", "b", "a" };

		std::array<MyData1, 6> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 6> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 7> MyCompare1
		const std::array<MyData1, 7> data{ "c", "a", "f", "g", "e", "d", "b" };
		const std::array<MyData1, 7> expected{ "g", "f", "e", "d", "c", "b", "a" };

		std::array<MyData1, 7> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 7> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData1, 8> MyCompare1
		const std::array<MyData1, 8> data{ "c", "a", "f", "h", "g", "e", "d", "b" };
		const std::array<MyData1, 8> expected{ "h", "g", "f", "e", "d", "c", "b", "a" };

		std::array<MyData1, 8> actual(data);
		fswn::sort(actual.begin(), actual.end(), MyCompare1{});

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData1, 8> MyCompare1 OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 0>
		const size_t N = 0;
		std::array<MyData2, N> data{};
		std::array<MyData2, N> actual{};
		fswn::sort(actual.begin(), actual.end());

		std::array<MyData2, N> expected{};
		std::stable_sort(expected.begin(), expected.end());

		//std::cout << data << std::endl;
		if (!AreContainerEqual(expected, actual)) {
			std::cout << "ERROR expected != actual" << std::endl;
			std::cout << "input = " << data << std::endl;
			std::cout << "expected = " << expected << std::endl;
			std::cout << "actual = " << actual << std::endl;
			return 1;
		}
		std::cout << "sort(first, last) std::array<MyData2, 0> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 1>
		const size_t N = 1;
		for (int v1 = 1; v1 <= N; ++v1) {
			const std::array<MyData2, N> data{
				MyData2{ v1, 1 }
			};

			std::array<MyData2, N> actual(data);
			fswn::sort(actual.begin(), actual.end());

			std::array<MyData2, N> expected(data);
			std::stable_sort(expected.begin(), expected.end());

			//std::cout << data << std::endl;
			if (!AreContainerEqual(expected, actual)) {
				std::cout << "ERROR expected != actual" << std::endl;
				std::cout << "input = " << data << std::endl;
				std::cout << "expected = " << expected << std::endl;
				std::cout << "actual = " << actual << std::endl;
				return 1;
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 1> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 2>
		const size_t N = 2;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				const std::array<MyData2, N> data{
					MyData2{ v1, 1 },
					MyData2{ v2, 2 }
				};

				std::array<MyData2, N> actual(data);
				fswn::sort(actual.begin(), actual.end());

				std::array<MyData2, N> expected(data);
				std::stable_sort(expected.begin(), expected.end());

				//std::cout << data << std::endl;
				if (!AreContainerEqual(expected, actual)) {
					std::cout << "ERROR expected != actual" << std::endl;
					std::cout << "input = " << data << std::endl;
					std::cout << "expected = " << expected << std::endl;
					std::cout << "actual = " << actual << std::endl;
					return 1;
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 2> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 3>
		const size_t N = 3;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					const std::array<MyData2, N> data{
						MyData2{ v1, 1 },
						MyData2{ v2, 2 },
						MyData2{ v3, 3 }
					};

					std::array<MyData2, N> actual(data);
					fswn::sort(actual.begin(), actual.end());

					std::array<MyData2, N> expected(data);
					std::stable_sort(expected.begin(), expected.end());

					//std::cout << data << std::endl;
					if (!AreContainerEqual(expected, actual)) {
						std::cout << "ERROR expected != actual" << std::endl;
						std::cout << "input = " << data << std::endl;
						std::cout << "expected = " << expected << std::endl;
						std::cout << "actual = " << actual << std::endl;
						return 1;
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 3> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 4>
		const size_t N = 4;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						const std::array<MyData2, N> data{
							MyData2{ v1, 1 },
							MyData2{ v2, 2 },
							MyData2{ v3, 3 },
							MyData2{ v4, 4 }
						};

						std::array<MyData2, N> actual(data);
						fswn::sort(actual.begin(), actual.end());

						std::array<MyData2, N> expected(data);
						std::stable_sort(expected.begin(), expected.end());

						//std::cout << data << std::endl;
						if (!AreContainerEqual(expected, actual)) {
							std::cout << "ERROR expected != actual" << std::endl;
							std::cout << "input = " << data << std::endl;
							std::cout << "expected = " << expected << std::endl;
							std::cout << "actual = " << actual << std::endl;
							return 1;
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 4> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 5>
		const size_t N = 5;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						for (int v5 = 1; v5 <= N; ++v5) {
							const std::array<MyData2, N> data{
								MyData2{ v1, 1 },
								MyData2{ v2, 2 },
								MyData2{ v3, 3 },
								MyData2{ v4, 4 },
								MyData2{ v5, 5 }
							};

							std::array<MyData2, N> actual(data);
							fswn::sort(actual.begin(), actual.end());

							std::array<MyData2, N> expected(data);
							std::stable_sort(expected.begin(), expected.end());

							//std::cout << data << std::endl;
							if (!AreContainerEqual(expected, actual)) {
								std::cout << "ERROR expected != actual" << std::endl;
								std::cout << "input = " << data << std::endl;
								std::cout << "expected = " << expected << std::endl;
								std::cout << "actual = " << actual << std::endl;
								return 1;
							}
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 5> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 6>
		const size_t N = 6;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						for (int v5 = 1; v5 <= N; ++v5) {
							for (int v6 = 1; v6 <= N; ++v6) {
								const std::array<MyData2, N> data{
									MyData2{ v1, 1 },
									MyData2{ v2, 2 },
									MyData2{ v3, 3 },
									MyData2{ v4, 4 },
									MyData2{ v5, 5 },
									MyData2{ v6, 6 }
								};

								std::array<MyData2, N> actual(data);
								fswn::sort(actual.begin(), actual.end());

								std::array<MyData2, N> expected(data);
								std::stable_sort(expected.begin(), expected.end());

								//std::cout << data << std::endl;
								if (!AreContainerEqual(expected, actual)) {
									std::cout << "ERROR expected != actual" << std::endl;
									std::cout << "input = " << data << std::endl;
									std::cout << "expected = " << expected << std::endl;
									std::cout << "actual = " << actual << std::endl;
									return 1;
								}
							}
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 6> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 7>
		const size_t N = 7;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						for (int v5 = 1; v5 <= N; ++v5) {
							for (int v6 = 1; v6 <= N; ++v6) {
								for (int v7 = 1; v7 <= N; ++v7) {
									const std::array<MyData2, N> data{
										MyData2{ v1, 1 },
										MyData2{ v2, 2 },
										MyData2{ v3, 3 },
										MyData2{ v4, 4 },
										MyData2{ v5, 5 },
										MyData2{ v6, 6 },
										MyData2{ v7, 7 }
									};

									std::array<MyData2, N> actual(data);
									fswn::sort(actual.begin(), actual.end());

									std::array<MyData2, N> expected(data);
									std::stable_sort(expected.begin(), expected.end());

									//std::cout << data << std::endl;
									if (!AreContainerEqual(expected, actual)) {
										std::cout << "ERROR expected != actual" << std::endl;
										std::cout << "input = " << data << std::endl;
										std::cout << "expected = " << expected << std::endl;
										std::cout << "actual = " << actual << std::endl;
										return 1;
									}
								}
							}
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 7> OK." << std::endl;
	} {
		// sort(first, last) std::array<MyData2, 8>
		const size_t N = 8;
		for (int v1 = 1; v1 <= N; ++v1) {
			for (int v2 = 1; v2 <= N; ++v2) {
				for (int v3 = 1; v3 <= N; ++v3) {
					for (int v4 = 1; v4 <= N; ++v4) {
						for (int v5 = 1; v5 <= N; ++v5) {
							for (int v6 = 1; v6 <= N; ++v6) {
								for (int v7 = 1; v7 <= N; ++v7) {
									for (int v8 = 1; v8 <= N; ++v8) {
										const std::array<MyData2, N> data{
											MyData2{ v1, 1 },
											MyData2{ v2, 2 },
											MyData2{ v3, 3 },
											MyData2{ v4, 4 },
											MyData2{ v5, 5 },
											MyData2{ v6, 6 },
											MyData2{ v7, 7 },
											MyData2{ v8, 8 }
										};

										std::array<MyData2, N> actual(data);
										fswn::sort(actual.begin(), actual.end());

										std::array<MyData2, N> expected(data);
										std::stable_sort(expected.begin(), expected.end());

										//std::cout << data << std::endl;
										if (!AreContainerEqual(expected, actual)) {
											std::cout << "ERROR expected != actual" << std::endl;
											std::cout << "input = " << data << std::endl;
											std::cout << "expected = " << expected << std::endl;
											std::cout << "actual = " << actual << std::endl;
											return 1;
										}
									}
								}
							}
						}
					}
				}
			}
		}
		std::cout << "sort(first, last) std::array<MyData2, 8> OK." << std::endl;
	} {
		using std::chrono::steady_clock;
		using std::chrono::microseconds;
		using std::chrono::duration_cast;

		// performance
		const int LOOP2 = 100000;
		const int LOOP3 = 100000;
		const int LOOP4 = 100000;
		const int LOOP5 = 10000;
		const int LOOP6 = 1000;
		const int LOOP7 = 100;
		const int LOOP8 = 10;
		{
			// std::sort() N = 2;
			const size_t N = 2;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP2; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						std::array<int, N> data{ v1, v2 };
						std::sort(data.begin(), data.end());
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP2 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 2;
			const size_t N = 2;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP2; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						std::array<int, N> data{ v1, v2 };
						fswn::sort(data.begin(), data.end());
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP2 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 3;
			const size_t N = 3;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP3; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							std::array<int, N> data{ v1, v2, v3 };
							std::sort(data.begin(), data.end());
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP3 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 3;
			const size_t N = 3;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP3; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							std::array<int, N> data{ v1, v2, v3 };
							fswn::sort(data.begin(), data.end());
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP3 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 4;
			const size_t N = 4;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP4; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								std::array<int, N> data{ v1, v2, v3, v4 };
								std::sort(data.begin(), data.end());
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP4 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 4;
			const size_t N = 4;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP4; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								std::array<int, N> data{ v1, v2, v3, v4 };
								fswn::sort(data.begin(), data.end());
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP4 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 5
			const size_t N = 5;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP5; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									std::array<int, 8> data{ v1, v2, v3, v4, v5 };
									std::sort(data.begin(), data.end());
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP5 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 5
			const size_t N = 5;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP5; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									std::array<int, 8> data{ v1, v2, v3, v4, v5 };
									fswn::sort(data.begin(), data.end());
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP5 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 6
			const size_t N = 6;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP6; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										std::array<int, 8> data{ v1, v2, v3, v4, v5, v6 };
										std::sort(data.begin(), data.end());
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP6 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 6
			const size_t N = 6;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP6; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										std::array<int, 8> data{ v1, v2, v3, v4, v5, v6 };
										fswn::sort(data.begin(), data.end());
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP6 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 7
			const size_t N = 7;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP7; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										for (int v7 = 1; v7 <= N; ++v7) {
											std::array<int, 8> data{ v1, v2, v3, v4, v5, v6, v7 };
											std::sort(data.begin(), data.end());
										}
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP7 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 7
			const size_t N = 7;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP7; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										for (int v7 = 1; v7 <= N; ++v7) {
											std::array<int, 8> data{ v1, v2, v3, v4, v5, v6, v7 };
											fswn::sort(data.begin(), data.end());
										}
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP7 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// std::sort() N = 8
			const size_t N = 8;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP8; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										for (int v7 = 1; v7 <= N; ++v7) {
											for (int v8 = 1; v8 <= N; ++v8) {
												std::array<int, 8> data{ v1, v2, v3, v4, v5, v6, v7, v8 };
												std::sort(data.begin(), data.end());
											}
										}
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "std::sort() N = " << N << " LOOP = " << LOOP8 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		} {
			// fswn::sort() N = 8
			const size_t N = 8;
			auto start = steady_clock::now();
			for (int loop = 0; loop < LOOP8; ++loop) {
				for (int v1 = 1; v1 <= N; ++v1) {
					for (int v2 = 1; v2 <= N; ++v2) {
						for (int v3 = 1; v3 <= N; ++v3) {
							for (int v4 = 1; v4 <= N; ++v4) {
								for (int v5 = 1; v5 <= N; ++v5) {
									for (int v6 = 1; v6 <= N; ++v6) {
										for (int v7 = 1; v7 <= N; ++v7) {
											for (int v8 = 1; v8 <= N; ++v8) {
												std::array<int, 8> data{ v1, v2, v3, v4, v5, v6, v7, v8 };
												fswn::sort(data.begin(), data.end());
											}
										}
									}
								}
							}
						}
					}
				}
			}
			auto end = steady_clock::now();
			std::cout << "fsw::sort() N = " << N << " LOOP = " << LOOP8 << " duration = " << duration_cast<microseconds>(end - start).count() / 1000000.0 << " [s]" << std::endl;
		}
	}

	return 0;
}
